# Brute force coherence (Gabriele Vajente, 2022-02-08)

import argparse
import os
 
from bruco import __version__

def parse_arguments():
    """
    Parse the command line arguments and return the values.

    Returns
    -------
    opt: class
        container for all the parsed parameters
    """

    helpstring = """
    Brute force coherence (Gabriele Vajente, 2022-02-03)
    
    Example:
    python -m bruco --ifo=H1 --channel=CAL-DELTAL_EXTERNAL_DQ 
               --calib=share/lho_cal_deltal_calibration.txt 
               --gpsb=1111224016 --length=600 --outfs=4096 --naver=100  
               --dir=./bruco_1111224016 --top=100 --webtop=20 --xlim=1:2048  
               --ylim=1e-21:1e-14 --excluded=share/lho_excluded_channels_O3.txt
    """
    
    ##### Define and get command line options ################################################
    parser = argparse.ArgumentParser(description=helpstring, 
                                     epilog="bruco version %s" % __version__, 
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    
    parser.add_argument("--channel", dest="channel",
                      default='', type=str, required=True,
                      help="name of the target channel")
    parser.add_argument("--file", dest="file",
                      default='', type=str,
                      help="specify an ASCII file containing the data to be " +\
                           "used as the main channel. " +\
                           "The --channel option will be ignored. ")
    parser.add_argument("--target_source", dest="target_source",
                      default="", type=str,
                      help="point to a file with a list of GWF files from where the target signal can be read")
    parser.add_argument("--aux_source", dest="aux_source",
                      default="", type=str,
                      help="point to a file with a list of GWF files from where the auxiliary signals can be read")
    parser.add_argument("--aux_num_read", dest="aux_num_read",
                      default=20, type=int,
                      help="maximum number of auxiliary channels to read simultaneously")
    parser.add_argument("--ifo", dest="ifo",
                      default="", type=str,
                      help="interferometer prefix [H1, L1]")
    parser.add_argument("--gpsb", dest="gpsb", required=True,
                      default=0, type=int,
                      help="start GPS time")
    parser.add_argument("--length", dest="dt",
                      default=0, type=int, required=True,
                      help="amount of data to use (in seconds)")
    parser.add_argument("--outfs", dest="outfs",
                      default=8192, type=int,
                      help="sampling frequency of the output results " + \
                           "(coherence will be computed up to outfs/2 " +\
                           "if possible)")
    parser.add_argument("--naver", dest="nav",
                      default=100, type=int,
                      help="number of FFTs to average")
    parser.add_argument("--fres", dest="fres",
                      default=0, type=float,
                      help="desired frequency resolution, overwrite the --naver parameter")
    parser.add_argument("--dir", dest="dir", type=str,
                      default='bruco_report',
                      help="output directory")
    parser.add_argument("--top", dest="ntop",
                      default=100, type=int,
                      help="number of top coherences saved (for each frequency) " +\
                           "in the datafiles idxtab.txt and cohtab.txt")
    parser.add_argument("--webtop", dest="wtop",
                      default=20, type=int,
                      help="number of top coherences written to the web page, for each frequency bin")
    parser.add_argument("--minfs", dest="minfs",
                      default=32, type=int,
                      help="minimum sampling frequency of aux channels, skip those with lower sampling rate")
    parser.add_argument("--plot", dest="plotformat",
                      default='png', type=str,
                      help="plot format (png, pdf or html)")
    parser.add_argument("--nproc", dest="ncpu",
                      default=-1, type=int,
                      help="number of processes to launch (if not specified, use all available cores)")
    parser.add_argument("--calib", dest="calib",
                      default='', type=str,
                      help="name of a text file containing the calibration "+\
                           "transfer function to be applied to the target "+\
                           "channel spectrum, in a two column format  " +\
                           "(frequency, absolute value)")
    parser.add_argument("--xlim", dest="xlim",
                      default='', type=str,
                      help="frequency axis limit, in the format fmin:fmax")
    parser.add_argument("--ylim", dest="ylim",
                      default='', type=str,
                      help="PSD y axis limits, in the format ymin:ymax")
    parser.add_argument("--excluded", dest="excluded",
                      default='', type=str,
                      help="point to a text file with a list of channels excluded from the coherence computation")
    parser.add_argument("--modulations", dest="modulations",
                      default='', type=str,
                      help="name of a file that containes the list of modulation channels")
    parser.add_argument("--select", dest="select",
                      default='', type=str,
                      help="wildcard used to select only a subset of auxiliary channels")

    opt = vars(parser.parse_args())

    # see if the user specified custom plot limits
    if opt['xlim'] != '':
        opt['xlim'] = [float(x) for x in opt['xlim'].split(':')]
    else:
        opt['xlim'] = [-1, -1]
    opt['xmin'], opt['xmax'] = opt['xlim']
    if opt['ylim'] != '':
        opt['ylim'] = [float(x) for x in opt['ylim'].split(':')]
    else:
        opt['ylim'] = [-1, -1]
    opt['ymin'], opt['ymax'] = opt['ylim']

    # extract the IFO identifier if not specified
    if opt['ifo'] == '':
        if opt['channel'][0:2] == 'H1' or opt['channel'][0:2] == 'L1':
            opt['ifo'] = opt['channel'][0:2]
            opt['channel'] = opt['channel'][3:]
        else:
            print('Error: IFO identifier not specified, and cannot detect from target channel name')
            exit()
    
    # tweak some parameters as needed
    opt['gpse'] = opt['gpsb'] + opt['dt']
    opt['dir'] = os.path.expanduser(opt['dir'])

    return opt



# Brute force coherence (Gabriele Vajente, 2022-02-08)

import numpy
import os
import multiprocessing
import pickle

from scipy.signal import resample_poly, resample, detrend
import scipy.stats

from . import __version__, __date__
from .functions import *
from .html import markup
from .parameters import *
from .ligodata import *
from .coherence import *
from .report import *


def bruco():
    """
    Main function, that is called when the module is execued from the command line
    """

    ## parse command line arguments
    opt = parse_arguments()

    print("********************************************************************************")
    print("****                    BruCo version %s (%s)                    ****" % (__version__, __date__))
    print("********************************************************************************")
    print()
    print("Analyzing data from gps %d to %d (%d s)." % (opt['gpsb'], opt['gpse'], opt['dt']))
    print()
 
    ## Find target data source ####################################################################
    if opt['target_source'] != '':
        # get list of files from file
        with open(opt['target_source'], 'r') as fid:
            opt['target_source'] = fid.read().splitlines()
    else:
        # by default we read the target channel with a custom function
        # that is able to identify the correct frame type
        opt['target_source'] = 'getTargetData'

    ## Find auxiliary data source  ################################################################
    if opt['aux_source'] != '':
        # get list of files from file
        with open(opt['aux_source'], 'r') as fid:
            opt['aux_source'] = fid.read().splitlines()
    else:
        # call a function to find raw data on disk
        opt['aux_source'] = find_data_path(opt['ifo'], opt['gpsb'], opt['gpse'])

    ## Read list of channels and keep only relevant channels ######################################
    channels, sample_rates = get_channel_list(opt['aux_source'])
    channels = select_channels(opt, channels, sample_rates)

    ## Read main channel data #####################################################################
    ch1, fs1 = read_target_data(opt)

    # check if the main channel is flat
    if ch1.min() == ch1.max():
        print("Error: main channel %s is flat!" % opt['channel'])
        exit()

    ## Resample to the output sampling frequency
    if opt['outfs'] == -1:
        # use the target channel sampling rate as analysis rate
        opt['outfs'] = fs1
    if fs1 > opt['outfs']:
        if fs1 % opt['outfs'] == 0:
            # the channel sampling rate is a multiple of the analysis rate
            ch1 = resample_poly(ch1, up=1, down=int(numpy.round(fs1) / opt['outfs']))
        else:
            # use a resample function that allows non integer ratios
            ch1 = resample(ch1, opt['outfs'], int(numpy.round(fs1)))
        fs1 = opt['outfs']

    ## Compute main channel FFT ###################################################################

    ## Decide the number of points for each FFT, setting it to the closer power of two
    if opt['fres'] != 0:
        # if the user specified a frequency resolution, try to match it
        npoints = opt['outfs']/opt['fres']
        opt['npoints'] = pow(2, int(numpy.log(npoints)/numpy.log(2)))
    else:
        opt['npoints'] = pow(2, int(numpy.log(2*(opt['gpse'] - opt['gpsb']) * \
                                     opt['outfs'] / opt['nav']) / numpy.log(2)))

    print("\nNumber of points            = %d" % opt['npoints'])
    dtime = opt['npoints']/opt['outfs']
    opt['naver'] = int(2*opt['dt']//dtime - 1)
    opt['fres'] = float(opt['outfs']) / float(opt['npoints'])
    print("Frequency resolution of FFT = %f Hz" % opt['fres'])
    print("Number of averages          = %d\n" % opt['naver'])

    ## Compute the main channels FFTs and PSD. Here I save the single segments FFTS,
    # to reuse them later on in the CSD computation. In this way, the main channel FFTs are
    # computed only once, instead of every iteration. This function returns the FFTs already
    # scaled is such a way that PSD = sum |FFT|^2, in units of 1/sqrt(Hz)
    ch1_ffts = computeFFTs(ch1, opt['npoints'], opt['npoints']//2, fs1)
    # compute the average to get the PSD
    psd1 = numpy.mean(numpy.abs(ch1_ffts)**2,1)
    # and compute the frequency bins
    f1 = numpy.linspace(0, fs1/2, int(opt['npoints']//2+1))

    ## Read the calibration transfer function, if specified
    if opt['calib'] != '':
        # load from file
        cdata = numpy.loadtxt(opt.calib)
        # interpolate to the right frequency bins
        opt['calib'] = numpy.interp(f1, cdata[:,0], cdata[:,1])
    else:
        # no calibration specified, use unity
        opt['calib'] = numpy.ones_like(f1)

    ## Compute the coherence confidence level based on the number of averages used in the PSD
    opt['s'] = scipy.stats.f.ppf(0.95, 2, 2*opt['nav'])
    opt['s'] = opt['s']/(opt['nav'] - 1 + opt['s'])

    ## Read modulation channels if any ############################################################

    if opt['modulations']:
        # read list of modulation channels
        modulations = open(opt['modulations'], 'r').read().splitlines()
        modulations = ifo_prefix(modulations, opt['ifo'])
        print('Reading modulation channels:\n    ' + '\n    '.join(modulations))
        # read data
        mod_data, mod_fs = getRawData(modulations, opt['gpsb'], opt['gpse'], opt['aux_source'])
        # create array of resampled data
        mod_channels = numpy.zeros((len(modulations)+1, (opt['gpse']-opt['gpsb'])*opt['outfs']))
        mod_channels[0,:] = 1   # including constant
        for i,m in enumerate(modulations):
            if mod_fs[m] < opt['outfs']:
                mod_channels[i+1,:] = resample_poly(mod_data[m], up=opt['outfs']//mod_fs[m], down=1)
            elif fs[m] > opt['outfs']:
                mod_channels[i+1,:] = resample_poly(mod_data[m], up=1, down=mod_fs[m]//opt['outfs'])
            else:
                mod_channels[i+1,:] = mod_data[m]
        mod_channels[i,:] = detrend(mod_channels[i,:])
        # add empty string to list of modulation channels, to represent unmodulated channel
        modulations = ['No modulation'] + modulations
    else:
        # there are no modulations
        modulations = ['No modulation']
        mod_channels = numpy.ones((1,(opt['gpse']-opt['gpsb'])*opt['outfs']))

    ## Start the multiprocess coherence computations ##############################################
        
    # split the list of channels in as many sublist as there are CPUs
    if opt['ncpu'] == -1:
        opt['ncpu'] = multiprocessing.cpu_count()

    # try the most even possible distribution of channels among the processes
    nchannels = len(channels)
    n = int(nchannels / opt['ncpu'])
    N1 = int( (nchannels / float(opt['ncpu']) - n) * opt['ncpu'])
    ch2 = []
    chidx = []
    for i in range(N1):
        ch2.append(channels[i*(n+1):(i+1)*(n+1)])
        chidx.append(i*(n+1))
    for i in range(opt['ncpu']-N1):
        ch2.append(channels[N1*(n+1)+i*n:N1*(n+1)+(i+1)*n])
        chidx.append(N1*(n+1)+i*n)

    # start a multiprocessing pool
    print(">>>>> Starting %d parallel processes..." % opt['ncpu'])
    if opt['ncpu'] > 1:
        pool = multiprocessing.Pool(opt['ncpu'])

    # Build the list of arguments
    args = []
    for i,c in enumerate(ch2):
        args.append([opt, ch1_ffts, f1, psd1, c, i, chidx[i], mod_channels, modulations])

    # Start all the processes
    if opt['ncpu'] > 1:
        results = pool.map(parallelized_coherence, args)
    else:
        results = [parallelized_coherence(args[0])]

    # when we get here, all the computations are concluded
    print(">>>>> Parallel processes finished...")

    ## Put all the results together ###############################################################

    cohtab, idxtab, modtab = collect_results(results, opt)

    ## Write the HTML report ######################################################################

    generate_report(opt, cohtab, idxtab, modtab, channels, modulations)


if __name__ == '__main__':
    # execute the main function when called from command line
    bruco()



# Brute force coherence (Gabriele Vajente, 2022-02-08)
# Auxiliary functions

import numpy
import scipy.signal
import fnmatch
import os
from gwpy.timeseries import TimeSeries

from .ligodata import *

def computeFFTs(x, npoints, noverlap, fs):
    """
    Compute the windowed FFTs of the channel x, each with npoints and overlapped by noverlap.

    Parameters
    ----------
    x: numpy.array
        channel time series
    npoints: int
        number of points used for each FFT
    noverlap: int
        number of points of overlap between FFTs
    fs: int
        channel sampling frequency

    Returns
    -------
    segs: numpy.array
        array of FFTs, normalized to gte PSD units when squared
    """

    # compute the starting indexes of all segments
    npoints = int(npoints)
    noverlap = int(noverlap)
    step = npoints - noverlap
    ind = numpy.arange(0, len(x) - npoints + 1, step, dtype=int)
    nsegs = len(ind)

    # pre-compute window
    wind = numpy.hanning(npoints)

    # pre-allocate output data
    segs = numpy.zeros((int(npoints/2+1), nsegs), dtype=numpy.cfloat)

    # compute all FFTs
    for i in range(nsegs):
        segs[:,i] = numpy.fft.rfft( scipy.signal.detrend(x[ind[i]:ind[i]+npoints], type='linear') * wind)

    # normalization to get PSD with |fft|^2
    segs = segs * numpy.sqrt(2.0/fs / numpy.sum(wind**2))

    # that's all
    return segs

def cohe_color(c):
    """
    Return an hex color code depending continuosly on input: 1 = red, 0 = white

    Parameters
    ----------
    c: float
        the input value between 0 and 1

    Returns
    -------
    color: str
        a hex color code varying between white and red
    """

    if c == 0:
        return '#ffffff'
    else:
        if c == 1:
            return '#ff0000'
        else:
            s = hex(int((1.0 - c) * 256)).split('x')[1]
            if len(s) == 1:
                s = '0' + s
            return '#ff' + s + s

def newline_name(s):
    """
    Split a channel name in multiple lines if it's too long.

    Parameters
    ----------
    s: str
        the channel name
    
    Returns
    -------
    s: str
        the same name, with inserted new lines
    """

    if len(s) > 10:
        N = int(len(s)/10)
        idx = []
        for i in range(N):
            try:
                idx.append(s.index('_', 10*(i+1)))
            except:
                pass
        if len(idx) !=0:
            idx  = numpy.unique(idx)
            newstr = ''
            for i in range(len(idx)):
                if i == 0:
                    newstr = s[0:idx[0]]
                else:
                    newstr = newstr + "<br>" + s[idx[i-1]:idx[i]]
            newstr = newstr + '<br>' + s[idx[-1]:]
            return newstr
        else:
            return s
    else:
        return s

def enc(x):
    """
    Some newer versions of matplotlib throw an error when labels have a '_'. So here I replace them with a blank.
    """
    return x.replace('_', ' ')

def select_channels(opt, channels, sample_rates):
    """
    Select only channels with high enough sampling rate and not in the exclusion list.

    Parameters
    ----------
    opt: dict
        parsed command line options
    channels: numpy.array of str
        list of all channels
    sample_rates: numpy.array of int
        list of corresponding sampling rates

    Returns
    -------
    channels: numpy.array of str
        the list of good channels
    sample_rates: numpy.array of int
        the list of corresponding sampling rates
    """

    # remove channels with low sampling rate
    idx = numpy.where(sample_rates >= opt['minfs'])[0]
    channels = channels[idx]
    sample_rates = sample_rates[idx]

    if opt['excluded'] != '':
        # load exclusion list from file
        with open(opt['excluded'], 'r') as f:
            L = f.readlines()
            excluded = []
            for c in L:
                c = c.split()
                if len(c):
                    excluded.append(c[0])
        opt['excluded'] = excluded
    
        # delete excluded channels, allowing for unix-shell-like wildcards
        idx = numpy.ones(channels.shape, dtype='bool')
        for i,c in enumerate(channels):
            if c == opt['ifo'] + ':' + opt['channel']:
                # remove the main channel
                idx[i] = False
            for e in excluded:
                if fnmatch.fnmatch(c, opt['ifo'] + ':' + e):
                    idx[i] = False

        channels = channels[idx]
        sample_rates = sample_rates[idx]

    # keep only channels that match the wildcards selection, if provided
    if opt['select'] != '':
        opt['select'] = opt['ifo'] + ':' + opt['select']
        print('Selecting only channels matching ' + opt['select'])
        idx = numpy.zeros(channels.shape, dtype='bool')
        for i,c in enumerate(channels):
            if fnmatch.fnmatch(c, opt['select']):
                idx[i] = True
        channels = channels[idx]
        sample_rates = sample_rates[idx]

    # remove repeated channels
    channels = numpy.unique(channels)

    # save the channel list to a file in the output directory
    os.makedirs(opt['dir'], exist_ok=True)
    with open(opt['dir'] + '/channels.txt', 'w') as fid:
        for c in channels:
            fid.write('%s\n' % c)
    opt['nch'] = channels.shape[0]

    print('Found %d auxiliary channels\n' % opt['nch'])

    return channels

def read_target_data(opt):
    """
    Read the target channel data.

    Parameters
    ----------
    opt: dict
        parsed command line arguments

    Returns
    -------
    ch1: numpy.array
        an array containing the channel data
    fs1: int
        channel sampling frequency
    """

    ## Read main channel data
    if opt['file'] == '':
        # read the main channel from frame data
        if opt['target_source'] == 'getTargetData':
            # no source specified, let gwpy find the data
            print('Reading target channel %s with gwpy' % (opt['ifo'] + ':' + opt['channel']))
            ch1, fs1 = getTargetData(opt['ifo'] + ':' + opt['channel'], opt['gpsb'], opt['gpse'])
        else:
            # read from the provided file list
            print('Reading target channel %s from files:' % (opt['ifo'] + ':' + opt['channel']))
            print('\n'.join(['    ' + t for t in opt['target_source']]))
            data = TimeSeries.read(opt['target_source'], opt['ifo'] + ':' + opt['channel'], 
                                   start=opt['gpsb'], end=opt['gpse'])
            ch1 = data.value
            fs1 = int(data.sample_rate.value)
    else:
        # the user provided a filename, read data from there
        print('Reading target channel %s from %s' % (opt['ifo'] + ':' + opt['channel'], opt['file']))
        extension = opt['file'].split('.')[-1]
        if extension == 'txt':
            ch1 = numpy.loadtxt(opt.file)
            fs1 = int(len(ch1)/(gpse-gpsb))
        elif extension == "gwf":
            x = TimeSeries.read(opt['file'], opt['ifo'] + ':' + opt['channel'], start=opt['gpsb'], end=opt['gpse'])
            ch1 = x.value
            fs1 = int(x.sample_rate.value)
        elif extension == 'pickle':
            import pickle
            x = pickle.load(open(opt['file'], 'rb'))
            ch1 = x[opt.channel]
            fs1 = int(len(ch1)/(gpse-gpsb))
        else:
            print('Error: file format not recognized for ' + opt['file'] +'. Must be txt, gwf or pickle')
            exit()
        print("    Read %d samples > sampling rate is %d" % (len(ch1), fs1))

    return ch1, fs1

def ifo_prefix(channels, ifo):
    """
    Make sure that the channels in the list start with the IFO prefix

    Parameters
    ----------
    channels: list of str
        list of channel names
    ifo: str
        IFO prefix ('H1' or 'L1')

    Returns
    -------
    channels: list of str
        list of channel names, all starting with the IFO prefix
    """
    new_channels = []
    for c in channels:
        if c[0:2] == ifo:
            new_channels.append(c)
        else:
            new_channels.append(ifo + ':' + c)
    return new_channels




# Brute force coherence (Gabriele Vajente, 2022-02-08)
# LIGO data access functions

import os
from urllib.parse import urlparse
import gwdatafind
from gwpy.timeseries import TimeSeries, TimeSeriesDict
import numpy

def find_data_path(observatory, gpsb, gpse):
    """
    Return the local disk path of the GWF files containing the RAW data.

    Parameters
    ----------
    observatory: str
        either 'H1' or 'L1'
    gpsb, gpse: int, int
        start and end GPS times of the desired RAW data

    Returns
    -------
    files: list of str
        list of filenames
    """

    obs = observatory[0]
    urls = gwdatafind.find_urls(
        obs,
        "{}1_R".format(obs),
        int(gpsb),
        int(gpse),
        urltype='file',
    )
    return [str(urlparse(x).path) for x in urls]

def get_channel_list(files):
    """
    Get the list of all channels that are available in the first file.

    Parameters
    ----------
    files: list of str
        list of filenames

    Returns
    -------
    channels: array of str
        channel names
    sample_rate: array of int
        corresponding sampling rates
    """

    print('Getting list of all channels from ' + files[0])

    # call external command to get the list of channels written to a file
    os.system('/usr/bin/FrChannels ' + files[0] + ' > bruco.channels')
    with open('bruco.channels', 'r') as f:
        lines = f.readlines()
    # parse all lines in the file
    channels = []
    sample_rate = []
    for l in lines:
        ll = l.split()
        if ll[0][1] != '0':
            # don't add L0/H0 channels, only L1 / H1 channels
            channels.append(ll[0])
            sample_rate.append(int(ll[1]))
    # convert to numpy arrays
    channels = numpy.array(channels)
    sample_rate = numpy.array(sample_rate)
    # remove temporary channel list file
    os.system('rm bruco.channels')

    return channels, sample_rate

def getRawData(channels, gpsb, gpse, files):
    """
    Read data from RAW files using gwpy

    Parameters
    ----------
    channels: list of str
        list of channel names
    gpsb, gpse: int, int
        start and end GPS times of desired data
    files: list of str
        list of GWF files where the data is available

    Returns
    -------
    data: dict
        dictionary containing the data, keys are channel names
    sampling_rate: dict
        dictionary containing the sampling rates, keys are channel names    

    """
    # read all channels using gwpy
    d = TimeSeriesDict.read(files, channels, start=gpsb, end=gpse)

    # extract data and sampling rate
    data = {}
    sampling_rate = {}

    for k in d.keys():
        data[k] = d[k].value
        sampling_rate[k] = int(d[k].sample_rate.value)

    return data, sampling_rate

def getTargetData(channel, gpsb, gpse):
    """
    Read target data using gwpy get() function

    Parameters
    ----------
    channel: str
        channel name
    gpsb, gpse: int, int
        start and end GPS times of desired data

    Returns
    -------
    data: numpy.array 
        array containing the data
    sampling_rate: int
        channel sampling rate    

    """
    # read channel using gwpy
    d = TimeSeries.get(channel, start=gpsb, end=gpse)

    # extract data and sampling rate
    data = d.value
    sampling_rate = int(d.sample_rate.value)

    return data, sampling_rate




# Brute force coherence (Gabriele Vajente, 2022-02-08)

import numpy
import os
import pickle
import gpstime
import sys
import time

from . import __version__, __date__
from .functions import *
from .html import markup

def collect_results(results, opt):
    """
    Collect all the results in single tables.

    Parameters
    ----------
    results: list
        list of results as returned by the multiprocessing computations
    opt: dict
        parsed command line arguments

    Returns
    -------
    cohtab: numpy.array
        table of all coherences, for all processes
    idxtab: numpy.array
        and table of corresponding channel indexes
    modstab: n umpy.array
        and table of corresponding modulation channel indexes
    """

    # first step, concatenate the tables
    x = list(zip(*results))
    cohtab = numpy.concatenate(x[0], axis=1)
    idxtab = numpy.concatenate(x[1], axis=1)
    modtab = numpy.concatenate(x[2], axis=1)

    # save timing info
    timing = x[4]
    with open(opt['dir'] + '/timing.pickle', 'wb') as timingfile:
        pickle.dump(timing, timingfile)

    # list of channels with errors
    errchannellist = numpy.concatenate(x[5],axis=0)
    numpy.savetxt(opt['dir']+'/channels_error.txt', errchannellist, fmt='%s')

    # then sort in order of descending coherence for each bin
    for j in numpy.arange(cohtab.shape[0]):
        ccoh = cohtab[j,:]
        iidx = idxtab[j,:]
        mmod = modtab[j,:]
        ii = ccoh.argsort()
        cohtab[j, :] = cohtab[j, ii]
        idxtab[j, :] = idxtab[j, ii]
        modtab[j, :] = modtab[j, ii] 
    # Finally, keep only the top values, which are the last columns
    cohtab = cohtab[:,-opt['ntop']:]
    idxtab = idxtab[:,-opt['ntop']:]
    modtab = modtab[:,-opt['ntop']:]
    # save the coherence tables to files
    numpy.savetxt(opt['dir'] + '/cohtab.txt', cohtab)
    numpy.savetxt(opt['dir'] + '/idxtab.txt', idxtab)
    numpy.savetxt(opt['dir'] + '/modtab.txt', modtab)

    return cohtab, idxtab, modtab

def generate_report(opt, cohtab, idxtab, modtab, channels, modulations):
    """
    Generate the HTML report page.

    Parameters
    ----------
    opt: dict
        parsed command line arguments
    cohtab: numpy.array
        table of all coherences, for all processes
    idxtab: numpy.array
        and table of corresponding channel indexes
    modtab: numpy.array
        and table of corresponding modulation indexes
    channels: list of str
        list of all channels
    modulations: list of str
        list of modulation names
    """

    dt = {'report': -time.time(), 'data': 0, 'decim':0, 'fft':0, 'plot':0, 'tot':0}

    print(">>>>> Generating report....")

    # get list of files, since they corresponds to the list of plots that have been created
    files = [f.split('.')[0] for f in os.listdir(opt['dir'])]

    # open web page
    page = markup.page( )
    page.init( title="Brute force Coherences",
               footer="(2022) <a href=mailto:vajente@caltech.edu>vajente@caltech.edu</a>" )
    page.h1('Top %d coherences of %s with auxiliary channels' % (opt['wtop'], opt['channel']))
    page.h2('GPS %d + %d s [%s]' % (opt['gpsb'], opt['dt'],
                  gpstime.tconvert(opt['gpsb'], form='%Y-%m-%d %H:%M:%S %Z')))

    # first section, top channels per frequency bin
    nf,nt = cohtab.shape
    freq  = numpy.linspace(0,opt['outfs']/2,nf)

    ## Write the main coherence table
    page.table(border=1, style='font-size:12px')
    page.tr()
    page.td(bgcolor="#5dadf1")
    page.h3('Frequency [Hz]')
    page.td.close()
    page.td(colspan=opt['ntop'], bgcolor="#5dadf1")
    page.h3('Top channels')
    page.td.close()
    page.tr.close()

    # here we create a huge table that contains, for each frequency bin, the list of most 
    # coherent channels, in descending order. The cell background color is coded from white 
    # (no coherence) to red (coherence 1)
    for i in range(nf):
        page.tr()
        page.td(bgcolor="#5dadf1")
        page.add("%.2f" % freq[i])
        page.td.close()
        for j in range(opt['wtop']):
            # write the channel only if the coherence in above the significance level
            if cohtab[i,-(j+1)] > opt['s']:
                page.td(bgcolor=cohe_color(cohtab[i,-(j+1)]))
                ch = (channels[int(idxtab[i,-(j+1)])]).split(':')[1]
                if modulations[int(modtab[i,-(j+1)])] != 'No modulation':
                    chm = ch  + ' * ' + modulations[int(modtab[i,-(j+1)])]
                else:
                    chm = ch
                if opt['plotformat'] != "none":
                    page.add("<a target=_blank href=%s.%s>%s</a><br>(%.2f)"
                             % (ch, opt['plotformat'], newline_name(chm), cohtab[i,-(j+1)]))
                else:
                    page.add("%s<br>(%.2f)" % (newline_name(chm), cohtab[i,-(j+1)]))
            else:
                page.td(bgcolor=cohe_color(0))
            page.td.close()
        page.tr.close()
    page.table.close()

    ## second section, links to all coherence plots
    if len(files)>0:
        page.h1('Coherence with all channels ')
        page.h2('GPS %d [%s] + %d s' % (opt['gpsb'],
                   gpstime.tconvert(opt['gpsb'], form='%Y-%m-%d %H:%M:%S %Z'), opt['dt']))

        N = len(files)
        m = 6     # number of channels per row
        M = int(N / m + 1)

        page.table(border=1)
        for i in range(M):
            page.tr()
            for j in range(m):
                if i*m+j < N:
                    page.td()
                    page.add('<a target=_blank href=%s.%s>%s</a>' % \
                                  (files[i*m+j], opt['plotformat'], files[i*m+j]))
                    page.td.close()
                else:
                    page.td()
                    page.td.close()
            page.tr.close()
        page.table.close()
        page.br()

    page.h2('Excluded channels:')
    page.code('  '.join(opt['excluded']))

    # Command call for reference
    page.h2('BruCo call string')
    page.code(' '.join(sys.argv))
    page.p()

    # That's the end, save the HTML page
    with open(opt['dir']  + '/index.html', 'w') as fileid:
        print(page, file=fileid)

    # timing info
    dt['report'] = dt['report'] + time.time()
    timing = pickle.load(open(opt['dir'] + '/timing.pickle', 'rb'))
    for t in timing:
        for k in t.keys():
            dt[k] += t[k]
    label = {'data':'Read data','decim':'Resampling','fft':'FFT','plot':'Plotting','tot':'Total', 'report':'Write report'}
    print()
    print('Task           Time  Time/nproc  Percent')
    print('----------------------------------------')
    for k in dt.keys():
        if k == 'report':
            print('%-12s          %6d' % (label[k], dt[k] ))
        elif k != 'tot':
            print('%-12s %6d   %6d      %3d' % (label[k], dt[k], dt[k]/opt['ncpu'], 100*dt[k]/dt['tot']))
    print('%-12s %6d   %6d' % (label['tot'], dt['tot'], dt['tot']/opt['ncpu']))




# Brute force coherence (Gabriele Vajente, 2022-02-08)

import numpy
import scipy.signal
import pylab
import matplotlib
import time
matplotlib.use("Agg")

from scipy.signal import resample_poly, resample, detrend

from .functions import *

def parallelized_coherence(args):
    """
    Compute the coherence of a subset of auxiliary channels with the target channel,
    and produce the coherence plots.

    Parameters
    ----------
    args: list
        a list of arguments, containing:
        opt: dict
            dictionary of command line arguments
        ch1_ffts: numpy.array
            precomputed FFTs of the target channel (ch1)
        f1: numpy.array
            corresponding frequency bins
        psd1: numpy.array
            averaged PSD of the target channel
        c: list of str
            list of channel names
        id: int
            process id number
        chidx: int
            channel index corresponding to the first in the list
        mod_channels: numpy.array
            resampled modulation channels
        modulations: list of str
            and corresponding channel names  
 
    Returns
    -------
    cohtab: numpy.array
        table of coherence values
    idxtab: numpy.array
        table of channel index values
    modtab: numpy.array
        table of modulation index values
    id: int
        process id number
    dt: dict
        timing information
    errchannels: list of str
        list fo the channels that had errors
    """

    # parse input arguments
    opt, ch1_ffts, f1, psd1, channels, id, chidx, mod_channels, modulations = args
    print("Called parallelized_coherence(), process %d" % id)

    # import locally the plotly library if needed and possible
    if opt['plotformat'] == 'html':
        try:
            from plotly.subplots import make_subplots
            import plotly.express as px
            import plotly.graph_objects as go
        except:
            try:
                import plotly
                print("WARNING: Your plotly version (" + plotly.__version__ + ") is too old, you need version 5.2.1 or later.")
            except:
                print("WARNING: You need the plotly library version 5.2.1 or later.")
            print("         Interactive plots disabled, saving plots in PNG format.")
            opt.plotformat = 'png'

    # dictionary to store timing information
    dt = {'data': 0, 'decim': 0, 'fft': 0, 'plot': 0, 'tot': 0}
    dt['tot'] = -time.time()

    # init tables
    cohtab = numpy.zeros((int(opt['npoints']/2+1), opt['ntop']))
    idxtab = numpy.zeros((int(opt['npoints']/2+1), opt['ntop']), dtype=int)
    modtab = numpy.zeros((int(opt['npoints']/2+1), opt['ntop']), dtype=int)
    # initialize empty list for the channels with errors
    errchannels = []

    # we might need multiple sampling frequencies for the modulation channels, so
    # we store them in a dictiornary once we need to compute them
    mod_channels = {opt['outfs']: mod_channels}

    # initialize figure and plot
    dt['plot'] = dt['plot'] - time.time()
    if opt['plotformat'] != 'none' and opt['plotformat'] != 'html':
        fig, ax = pylab.subplots(2, 1, sharex=True, figsize=(10,10))
        firstplot = True
    if opt['plotformat'] != 'none':
        # calibrated PSD of target channel, for plotting
        psd_plot = numpy.sqrt(psd1) * opt['calib']
    dt['plot'] = dt['plot'] + time.time()

    # analyze every channel in the list, reading them in batches to speed-up data IO
    nchannels = len(channels)
    channels_id = numpy.arange(nchannels)
    # divide in batches
    channels    = [channels[i:i+opt['aux_num_read']] for i in range(0, nchannels, opt['aux_num_read'])]
    channels_id = [channels_id[i:i+opt['aux_num_read']] for i in range(0, nchannels, opt['aux_num_read'])]
    # count progress
    count = 0

    for channels_subset, channels_id_subset in zip(channels, channels_id):
        # read all channels in this subset
        dt['data'] = dt['data'] - time.time()
        print("  Process %d: reading\n        %s" % (id, '\n        '.join(channels_subset)))
        data, sampling_rates = getRawData(channels_subset, opt['gpsb'], opt['gpse'], opt['aux_source'])
        dt['data'] = dt['data'] + time.time()

        # loop over all the channels we just got
        for i,channel2 in zip(channels_id_subset, channels_subset):
            count = count + 1
            print("  Process %d: channel %d of %d: %s" % \
                     (id, count, nchannels, channel2))
            ch2 = data[channel2]
            fs2 = sampling_rates[channel2]
            # check if the channel is flat, and skip it if so
            if ch2.min() == ch2.max():
                print("  Process %d: %s is flat, skipping" % (id, channel2))
                continue

            # resample to outfs if needed
            dt['decim'] = dt['decim'] - time.time()
            if fs2 > opt['outfs']:
                if fs2 % opt['outfs'] == 0:
                    ch2 = resample_poly(ch2, up=1, down=int(numpy.round(fs2) / opt['outfs']))
                else:
                    ch2 = resample(ch2, opt['outfs'], numpy.round(fs2))
                fs2 = opt['outfs']
            elif fs2 < opt['outfs']:
                # when the sampling frequency of the channel is smaller than outfs
                # we need to resample the modulation channels
                if fs2 not in list(mod_channels.keys()):
                    mod_channels[fs2] = resample_poly(mod_channels[opt['outfs']], axis=1, 
                                              up=1, down=int(opt['outfs']//fs2))
            dt['decim'] = dt['decim'] + time.time()

            ch2 = detrend(ch2)

            ## loop over all modulations, and keep all computed coherences
            C = []
            for k,modul in enumerate(modulations):
                # compute coherence
                dt['fft'] = dt['fft'] - time.time()
                # compute all the FFTs of the aux channel (those FFTs are returned already 
                # normalized so that the PSD is just sum |FFT|^2)
                # we use a number of points so that the frequency binning is the same as the
                # target channel FFTs
                ch2_ffts = computeFFTs(ch2*mod_channels[fs2][k,:], int(opt['npoints']*fs2/opt['outfs']), 
                                        int(opt['npoints']*fs2/opt['outfs']/2), fs2)
                # average to get PSDs and CSDs, create frequency vector
                psd2 = numpy.mean(abs(ch2_ffts)**2, axis=1)
                f = numpy.linspace(0, int(fs2/2), int(opt['npoints']*fs2/opt['outfs']/2+1))
                csd12 = numpy.mean(numpy.conjugate(ch2_ffts)*\
                               ch1_ffts[0:int(opt['npoints']*fs2/opt['outfs']/2+1),:], axis=1)
                # we use the full sampling PSD of the main channel, using only the bins 
                # corresponding to channel2 frequency bins
                c = numpy.abs(csd12)**2/(psd2 * psd1[0:len(psd2)])
                # mask frequencies above Nyquist
                c[f > fs2/2] = 0
                # save coherecne to list
                C.append(c)
                dt['fft'] = dt['fft'] + time.time()

                # save coherence in summary table. Basically, cohtab has a row for each frequency 
                # bin and a number of columns which is determined by the option --top. For each 
                # frequency bin, the new coherence value is added only if it's larger than the 
                # minimum already present. Then idxtab contains again a row for each frequency
                # bin: but in this case each entry is an unique index that determines the channel 
                # that gave that coherence. Finally modtab contains the corresponding modulation
                # channel index, if modulations are present. Modulation index 0 means no modulation.
                # So for example cohtab[100,0] gives the highest coherence for the 100th frequency 
                # bin; idxtab[100,0] contains an integer id that corresponds to the channel. This 
                # id is saved in channels.txt. modtab[100,0] contains the corresponging modulation
                # channel index.
                for j,cx in enumerate(c):
                    top = cohtab[j, :]
                    idx = idxtab[j, :]
                    mod = modtab[j, :]
                    # if the coherence of this channel at this frequency is higher than the smaller
                    # on record, then we add it and keep only the highest values and channels
                    if cx > top.min():
                        ttop = numpy.concatenate((top, [cx]))         # new coherence value
                        iidx = numpy.concatenate((idx, [chidx + i]))  # and corresponding new channel
                        mmod = numpy.concatenate((mod, [k]))          # and modulation channel
                        ii = ttop.argsort()
                        ii = ii[1:]
                        cohtab[j, :] = ttop[ii]
                        idxtab[j, :] = iidx[ii]
                        modtab[j, :] = mmod[ii]

            # create the output plot, if desired, and with the desired format
            dt['plot'] = dt['plot'] - time.time()
            if opt['plotformat'] != "none" and opt['plotformat'] != 'html':
                ## Use matplotlib for png or pdf static plot

                mask = numpy.ones_like(f)
                mask[c < opt['s']] = numpy.nan
                # faster plotting, create all the figure and axis stuff once for all,
                # at the beginning, the just upadte the traces and texts
                if firstplot:
                    pltitle = ax[0].set_title('Coherence %s vs %s - GPS %d' % \
                                        (enc(opt['channel']), enc(channel2), opt['gpsb']), \
                                         fontsize='smaller')
                    # make a list of non-repeating colors and line styles
                    colors = list(matplotlib.colors.TABLEAU_COLORS) + list(matplotlib.colors.BASE_COLORS)[:-2]
                    line_styles = ['-']*len(colors) + [':']*len(colors)
                    line_colors = colors + colors
                    line1 = {}
                    line4 = {}
                    line3, = ax[1].loglog(f1, psd_plot[0:len(f1)], color='k', alpha=1, label='Target')
                    for k,m in enumerate(modulations):
                        if len(modulations) > 1:
                            line1[k], = ax[0].loglog(f, C[k], label=m, alpha=1, linestyle=line_styles[k], color=line_colors[k])
                        else:
                            line1[k], = ax[0].loglog(f, C[k], label='Coherence', alpha=1, linestyle=line_styles[k], color=line_colors[k])
                        mask = numpy.ones_like(f)
                        mask[C[k]<opt['s']] = numpy.nan
                        if len(modulations) > 1:
                            line4[k], = ax[1].loglog(f, psd_plot[0:len(psd2)] * numpy.sqrt(C[k]) * mask, label=m, alpha=1,
                                                                         linestyle=line_styles[k], color=line_colors[k])
                        else:
                            line4[k], = ax[1].loglog(f, psd_plot[0:len(psd2)] * numpy.sqrt(C[k]) * mask, label='Projection', alpha=1,
                                                                         linestyle=line_styles[k], color=line_colors[k])
                    line2, = ax[0].loglog(f, numpy.ones_like(f)*opt['s'], 'r--', linewidth=1)
                    if opt['xmin'] != -1:
                        ax[0].axis(xmin=opt['xmin'], xmax=opt['xmax'])
                        ax[1].axis(xmin=opt['xmin'], xmax=opt['xmax'])
                    else:
                        ax[0].axis(xmax=opt['outfs']/2)
                    ax[0].axis(ymin=opt['s']/2, ymax=1)
                    ax[0].grid(True, alpha=0.5, linestyle=':')
                    ax[0].set_ylabel('Coherence')

                    if opt['ymin'] != -1:
                        ax[1].axis(ymin=opt['ymin'], ymax=opt['ymax'])
                    ax[1].set_xlabel('Frequency [Hz]')
                    ax[1].set_ylabel('Spectrum')
                    ax[1].legend(('Target channel', 'Noise projection'))
                    ax[1].grid(True, alpha=0.5, linestyle=':')
                    pylab.subplots_adjust(bottom=0.1, left=0.05, right=0.75, top=0.85, hspace=0.15)
                    firstplot = False
                else:
                    # if not the first plot, just update the traces and title
                    pltitle.set_text('Coherence %s vs %s - GPS %d' % (enc(opt['channel']), enc(channel2), opt['gpsb']))
                    for k,m in enumerate(modulations):
                        line1[k].set_data(f, C[k])
                        mask = numpy.ones_like(f)
                        mask[C[k]<opt['s']] = numpy.nan
                        line4[k].set_data(f, psd_plot[0:len(psd2)] * numpy.sqrt(C[k]) * mask)
                # add legends
                if len(modulations) > 1:
                    ax[0].legend(fontsize=8, title='Modulation', bbox_to_anchor=(1.05, 1.0), loc='upper left')
                    ax[1].legend(fontsize=8, title='Modulation', bbox_to_anchor=(1.05, 1.0), loc='upper left')
                else:
                    ax[1].legend(fontsize=8, bbox_to_anchor=(1.05, 1.0), loc='upper left')
                pylab.tight_layout()

                # save figure to file
                fig.savefig(opt['dir'] + '/%s.%s' % (channel2.split(':')[1], opt['plotformat']), format=opt['plotformat'])
            elif opt['plotformat'] == 'html':
                ## use plotly for interactive HTML plot

                fig = make_subplots(rows=2, cols=1, shared_xaxes=True, vertical_spacing=0.02)

                for k,m in enumerate(modulations):
                    if len(modulations) > 1:
                        fig.add_trace(go.Scatter(x=f, y=C[k], mode="lines", name=m,
                                             line_color=px.colors.qualitative.Alphabet[k],
                                             legendgroup=m, hovertemplate='freq: %{x:.2f}<br>coher: %{y:.4f}'), row=1, col=1)
                    else:
                        fig.add_trace(go.Scatter(x=f, y=C[k], mode="lines", name='Coherence',
                                             line_color=px.colors.qualitative.Alphabet[k], 
                                             legendgroup=m, hovertemplate='freq: %{x:.2f}<br>coher: %{y:.4f}'), row=1, col=1)
                    mask = numpy.ones_like(f)
                    mask[C[k]<opt['s']] = numpy.nan
                    if len(modulations) > 1:
                        fig.add_trace(go.Scatter(x=f, y=psd_plot[0:len(psd2)] * numpy.sqrt(C[k]) * mask, mode="lines",
                                                  line_color=px.colors.qualitative.Alphabet[k], name=m,
                                                  legendgroup=m, showlegend=False,
                                                  hovertemplate='freq: %{x:.2f} Hz<br>ASD: %{y:.4f}'), row=2, col=1)
                    else:
                        fig.add_trace(go.Scatter(x=f, y=psd_plot[0:len(psd2)] * numpy.sqrt(C[k]) * mask, mode="lines",
                                                  line_color=px.colors.qualitative.Alphabet[k], name='Projection',
                                                  legendgroup=m, showlegend=False,
                                                  hovertemplate='freq: %{x:.2f} Hz<br>ASD: %{y:.4f}'), row=2, col=1)
                fig.add_trace(go.Scatter(x=f, y=numpy.ones_like(f)*opt['s'], mode="lines", name='Threshold', line_color='red',
                                                       legendgroup='1', line={'dash':'dot'},
                                                  hovertemplate='freq: %{x:.2f} Hz<br>coher: %{y:.4f}'), row=1, col=1)
                fig.add_trace(go.Scatter(x=f, y=psd_plot[0:len(f1)], mode="lines", name='Target signal',
                                                       legendgroup='2', line_color='black',
                                                  hovertemplate='freq: %{x:.2f} Hz<br>ASD: %{y:.4f}'), row=2, col=1)

                fig.update_xaxes(type="log")
                fig.update_yaxes(type="log", exponentformat='power')
                fig.update_layout(title='Coherence %s vs %s - GPS %d' % (enc(opt['channel']), enc(channel2), opt['gpsb']))
                fig.update_layout(legend_tracegroupgap=5, template='plotly')
                fig.update_xaxes(title_text='Frequency [Hz]', row=2, col=1)
                fig.update_yaxes(title_text='Coherence', row=1, col=1)
                fig.update_yaxes(title_text='Spectrum', row=2, col=1)
                fig.update_layout(hoverlabel=dict(bgcolor="white", font_size=10))

                if opt['xmin'] != -1:
                    fig.update_xaxes(range=[numpy.log10(opt['xmin']),numpy.log10(opt['xmax'])])
                else:
                    fig.update_xaxes(range=[-1,numpy.log10(opt['outfs']/2)])
                fig.update_yaxes(range=[numpy.log10(opt['s']/2),0], row=1, col=1)
                if opt['ymin'] != -1:
                    fig.update_yaxes(range=[numpy.log10(opt['ymin']),numpy.log10(opt['ymax'])], row=2, col=1)

                fig.write_html(opt['dir'] + '/%s.%s' % (channel2.split(':')[1], opt['plotformat']), include_plotlyjs='cdn')

            dt['plot'] = dt['plot'] + time.time()
            # do some cleanup of memory
            del ch2, c, f
    # close figure, clean-up
    del fig

    print("  Process %s concluded" % id)
    dt['tot'] = dt['tot'] + time.time()
    return cohtab, idxtab, modtab, id, dt, errchannels





# -*- utf-8 -*-
# Copyright 2018 Gabriele Vajente

"""Install script for Bruco, install with pip:

    python -m pip install .
"""

import glob
import os.path
import re

from setuptools import (setup, find_packages)


def find_version(path):
    """Parse the __version__ metadata in the given file.
    """
    with open(path, 'r') as fp:
        version_file = fp.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    # metadata
    name='bruco',
    provides=['bruco'],
    description="Brute force coherence",
    author="Gabriele Vajente",
    license="GPLv3+",
    url="https://git.ligo.org/gabriele-vajente/bruco",
    version=find_version(os.path.join('bruco', '__init__.py')),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: '
            'GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    # dependencies
    setup_requires=[
        'setuptools',
    ],
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'gwdatafind',
        'gwpy',
    ],
    # content
    packages=find_packages(),
    scripts=[os.path.join('bin', 'bruco')],
    data_files=[
        (os.path.join('share', 'bruco'),
         list(glob.glob(os.path.join('share', '*.txt'))),
         ),
    ],
)
